/* The core of the pyifp system. This interface description generates the Python 
   extension module.
   
   $Id: ifp_core.i,v 1.7 2005/02/21 03:01:40 milamber Exp $
   
*/

%feature("autodoc", "1"); 
/* This requires a version of SWIG patched with Robert Dunn's patches from wxPython to work */

%define DOCSTRING
"A thin wrapper around the libifp library"
%enddef

%module(docstring=DOCSTRING) ifp_core
%module ifp_core

%include "typemaps.i"


/* I separated out the usb stuff to keep myself from getting confused */
%import "usb_core.i"

/*	Constants and Structs from ifp.h
	These make it easier to use these from python
*/
#define IFP_FILE   1
#define IFP_DIR   2
#define IFP_ERR_DEV_FUBAR   8
#define IFP_ERR_BAD_FREQUENCY   9
#define IFP_ERR_BAD_FILENAME   10
#define IFP_ERR_USER_CANCEL   11
#define IFP_TUNER_PRESET_DATA   240
#define IFP_FREQ_MIN   8750
#define IFP_FREQ_MAX   10800
#define IFP_PRESET_TOTAL   20
#define IFP_TUNER_LABEL   6

#define IFP_MAX_PATH		0x80
#define IFP_MAX_FILENAME	0x80

enum {IFP_WALK_FILE = IFP_FILE, IFP_WALK_DIR_PRE = IFP_DIR, IFP_WALK_DIR_POST, IFP_WALK_NONE};

/* Placing this here lets me access the fields of the struct from python */
struct ifp_transfer_status {
	int file_bytes; /**< the number of bytes transferred for the current file */
	int file_total; /**< the expected number of bytes in the current file */
	%immutable;
	const char * file_name; /**< the name of the current file */
	%mutable;
	long batch_bytes; /**< The number of bytes transferred in the transfer so far */
	long batch_total; /**< the total number of bytes expected during this transfer */

	int files_count; /**< The number of files successfully transferred so far. */
	int files_total; /**< The number of files expected to be transferred in this batch. */

	int is_batch; /**< 0 during single-file transfers, 1 during multi-file transfers */
};

%inline %{
typedef struct ifp_device ifp_device;
%}

int ifp_init(struct ifp_device *ifp_device,void *dev_handle);

%{
	#include <usb.h>
	#include <ifp.h>


	/* I can't figure out how to allocate and pass a buffer and the buffer's max 
	   size from python to c, so I'll just wrap the functions that needed it in a
	   little bit of C
	*/
	char *get_ifp_model(struct ifp_device *ifp_dev) {
		char *buffer;
		int i;
		
		buffer = malloc(255);
		i = ifp_model(ifp_dev,buffer,255);
		if(i != 0) {
			return NULL;
		}
		
		return buffer;
	}
	
	/* return the human readable info string from the iFP device */
	char *get_ifp_device_info(struct ifp_device *ifp_dev) {
		char *buffer;
		int i;
		buffer = (char *)malloc(255);
		i = ifp_device_info(ifp_dev,buffer,255);
		if(i != 0) {
			return NULL;
		}
		return buffer;
	}
	
	
	/* Functions to handle radio station presets */
	/* Get the information for a single preset radio station
	
	   It might be better to keep a refrence to the preset data around between calls (even giving it to python
	   to manage) but this is simpler and seems to be fast enough even setting and reading all 20 presets.
	*/
	int get_station_info(struct ifp_device *ifp_dev,int station_num,char *callsign, int *freq) {
		void *preset_data;
		int res;
		
		preset_data = malloc(IFP_TUNER_PRESET_DATA);
		
		res = ifp_get_tuner_presets(ifp_dev,preset_data,IFP_TUNER_PRESET_DATA);
		if(res != 0) {
			free(preset_data);
			return res;
		}
		
		res = ifp_get_station(station_num,preset_data,callsign,freq);
		free(preset_data);	
		return res;
	}
	
	/* Set the information for a single preset radio station
	
	   It might be better to keep a refrence to the preset data around between calls (even giving it to python
	   to manage) but this is simpler and seems to be fast enough even setting and reading all 20 presets.
	   
	   TODO: some length checking on the callsign should go in here. Setting the callsign to 7 chars 
	         seems to work fine in my test, but that may be chance.
	*/
	int set_station_info(struct ifp_device *ifp_dev,int station_num,const char *callsign, int freq) {
		void *preset_data;
		int res;
		
		/* check the validity of the preset data */
		if(strlen(callsign) > IFP_TUNER_LABEL) {
			return -EINVAL;
		}
		if(freq < IFP_FREQ_MIN || freq > IFP_FREQ_MAX) {
			return -EINVAL;
		}
		preset_data = malloc(IFP_TUNER_PRESET_DATA);
		
		/* retrieve the current preset data from the iFP device */
		res = ifp_get_tuner_presets(ifp_dev,preset_data,IFP_TUNER_PRESET_DATA);
		if(res != 0) {
			free(preset_data);
			return res;
		}
		
		res = ifp_set_station(station_num,preset_data,callsign,freq);
		if(res != 0) {
			free(preset_data);
			return res;
		}
		
		res = ifp_set_tuner_presets(ifp_dev,preset_data,IFP_TUNER_PRESET_DATA);
		free(preset_data);
		return res;
	}
	
	/* Callback function to wrap a call to a python callback function.
	   The python callable object (with signature [self,int,int,string]) is passed in
	   as the callback's "context." I've renamed the paramater to make it clean what it should be
	*/
	int list_dir_callback(void *python_callback,int type,const char* filename,int filesize) {
		PyObject *args;
		PyObject *result = NULL;
		int i;
		
		args = Py_BuildValue("iis",type,filesize,filename);
		result = PyObject_CallObject((PyObject *)python_callback,args);
		Py_DECREF(args);
		if (result == NULL || result == Py_None) { // I think None == NULL here
			return 0;
		}
		
		i = (int)PyInt_AsLong(result);
		Py_XDECREF(result);
		
		return i;
	}
	
	/* Call the given python method for each file/directory in the named directory.
	   To make this more pythonic it really should return a tuple of tuples for the directory.
	   Of course, the python callback can create this, but it would mean some static data needs
	   be kept between callbacks, which doesn't feel like a good thing.
	   
	   The void *context argument has been replaced by PyObject *py_callback to help me keep track
	   of all the callbacks going on in the code.
	   
	   TODO: If the py_callback is NULL then it could return a tuple of tuples in the context.
	         I'm not sure how easy that would be to wrap though.
	*/
	int list_dir(struct ifp_device *ifp_dev,const char *filename,PyObject *py_callback) {
		int i;
		
		
		Py_INCREF(py_callback);
		i = ifp_list_dirs(ifp_dev,filename,&list_dir_callback,py_callback);
		
		Py_XDECREF(py_callback);
		return i;
	}

	/* Translate the callback from the library into a call to a python method.
	
	   Wrapping the callback is the same no matter what direction you're going
	*/
	int up_down_load_file_py_wrap(void *context,struct ifp_transfer_status *status) {
		
		PyObject *statusObject = SWIG_NewPointerObj(status,SWIGTYPE_p_ifp_transfer_status,0);
		
		PyObject *args = Py_BuildValue("(O)",statusObject);
		PyObject *result;
		int i;
		
		result = PyEval_CallObject((PyObject*)context,args);
		Py_DECREF(args);
		Py_DECREF(statusObject);
		if(result == NULL || result == Py_None) {
			i = 0;
		} else {
			i = (int)PyInt_AsLong(result);
		}
		Py_DECREF(result);
		return i;

	}
	
	/* upload a named file to the iFP with a python callback.
	   The python callable needs to have the signature:
	   (transfer_status)
	   
	   The void *context object has been renamed PyObject *py_callback to make what is going on clearer.
	   
	   The error to Exception translation should probably be broken out into a %exception handler
	*/
	int upload_file_status(struct ifp_device *ifp_dev,const char *localfile, const char *remotefile,PyObject *py_callback) {
		int i;

		if(!PyCallable_Check(py_callback)) {
			return -EINVAL;
		}
		Py_INCREF(py_callback);
		i = ifp_upload_file(ifp_dev,localfile,remotefile,&up_down_load_file_py_wrap,py_callback);
		Py_DECREF(py_callback);
		
		return i;
	}
	
	/* download a named file from the iFP with a python callback.
	   The python callable needs to have the signature:
	   (transfer_status)
	   
	   The void *context object has been renamed PyObject *py_callback to make what is going on clearer.
	   
	   The error to Exception translation should probably be broken out into a %exception handler
	*/
	int download_file_status(struct ifp_device *ifp_dev,const char *remotefile, const char *localfile,PyObject *py_callback) {
		int i;

		if(!PyCallable_Check(py_callback)) {
			return -EINVAL;
		}
		Py_INCREF(py_callback);
		i = ifp_download_file(ifp_dev,remotefile,localfile,&up_down_load_file_py_wrap,py_callback);
		Py_DECREF(py_callback);
		
		return i;
	}
	
	/* upload a diretory to the iFP with a python callback.
	   The python callable needs to have the signature:
	   (transfer_status)
	   
	   The void *context object has been renamed PyObject *py_callback to make what is going on clearer.
	   
	   The error to Exception translation should probably be broken out into a %exception handler
	*/
	int upload_dir_status(struct ifp_device *ifp_dev,const char *localdir, const char *remotedir,PyObject *py_callback) {
		int i;

		if(!PyCallable_Check(py_callback)) {
			return -EINVAL;
		}
		Py_INCREF(py_callback);
		i = ifp_upload_dir(ifp_dev,localdir,remotedir,&up_down_load_file_py_wrap,py_callback);
		Py_DECREF(py_callback);
		
		return i;
	}
	
	/* download a diretory from the iFP with a python callback.
	   The python callable needs to have the signature:
	   (transfer_status)
	   
	   The void *context object has been renamed PyObject *py_callback to make what is going on clearer.
	   
	   The error to Exception translation should probably be broken out into a %exception handler
	*/
	int download_dir_status(struct ifp_device *ifp_dev,const char *remotedir, const char *localdir,PyObject *py_callback) {
		int i;

		if(!PyCallable_Check(py_callback)) {
			return -EINVAL;
		}
		Py_INCREF(py_callback);
		i = ifp_download_dir(ifp_dev,remotedir,localdir,&up_down_load_file_py_wrap,py_callback);
		Py_DECREF(py_callback);
		
		return i;
	}
%}


%exception list_dir {
	if(!PyCallable_Check(arg3)) {
		PyErr_SetString(PyExc_TypeError,"Callback must be a callable object");
		goto fail;
	}
	$function
	if(result) {
		PyErr_SetString(PyExc_IOError,ifp_error_message(result));
		goto fail;
	}
}
		
int list_dir(struct ifp_device *ifp_dev,const char *filename, PyObject *callback);


/* Get the firmware revision in BCD format. 
	From python: firmware_version = "%x.%02x" % (i / 0x0100,i % 0x0100)
*/
int ifp_firmware_version(struct ifp_device *ifp_dev);

/* can't import this at the top or I get a syntax error in cmalloc.i on line 24 in the middle of a comment!?!*/
%import "cmalloc.i"
/* While I wrap malloc for struct ifp_device, the free takes place in an exception handler for ifp_finalize */
%malloc(ifp_device);

/* Named exception handlers to return custom error messages */
%exception ifp_find_device {
	$function
	if(result == NULL) {
		PyErr_SetString(PyExc_IOError,"A suitable iRiver iFP device couldn't be found. Perhaps it's unpluged or turned off.");
		goto fail;
	}
}

/* This isn't actually an exception handler, it just cleans up by freeing the device handle
   after the call to ifp_finalize.
*/
%exception ifp_finalize {
	$function
	free(arg1);
}

%exception get_ifp_model {
	$function
	if(result == NULL) {
		PyErr_SetString(PyExc_IOError,"Unable to retrieve the model string from the iFP device.");
		goto fail;
	}
}

%exception get_ifp_device_info {
	$function
	if(result == NULL) {
		PyErr_SetString(PyExc_IOError,"Unable to retrieve the device info string from the iFP device.");
		goto fail;
	}
}

%exception upload_file_status {
	if(!PyCallable_Check(arg4)) {
			PyErr_SetString(PyExc_TypeError,"Callback must be a callable object");
			goto fail;
	}
	$function
	switch(result) {
			case 0:
				break;
			case -EEXIST:
				PyErr_SetString(PyExc_IOError,"File already exists");
				break;
			case -ENOENT:
				PyErr_SetString(PyExc_IOError,"File or directory does not exist");
				break;
			case -ENOSPC:
				PyErr_SetString(PyExc_IOError,"No space left on device");
				break;
			case IFP_ERR_BAD_FILENAME:
				PyErr_SetString(PyExc_IOError,"Bad filename");
				break;
			case IFP_ERR_USER_CANCEL: /* Should this really be an error?*/
				PyErr_SetString(PyExc_IOError,"User canceled operation");
				break;
			default:
				PyErr_SetString(PyExc_IOError,"Unknown error occured");
				break;
		}
		if(result)
			goto fail;
}

%exception download_file_status {
	if(!PyCallable_Check(arg4)) {
			PyErr_SetString(PyExc_TypeError,"Callback must be a callable object");
			goto fail;
	}
	$function
	switch(result) {
			case 0:
				break;
			case -EEXIST:
				PyErr_SetString(PyExc_IOError,"File already exists");
				break;
			case -ENOENT:
				PyErr_SetString(PyExc_IOError,"File or directory does not exist");
				break;
			case -ENOSPC:
				PyErr_SetString(PyExc_IOError,"No space left on device");
				break;
			case IFP_ERR_BAD_FILENAME:
				PyErr_SetString(PyExc_IOError,"Bad filename");
				break;
			case IFP_ERR_USER_CANCEL: /* Should this really be an error?*/
				PyErr_SetString(PyExc_IOError,"User canceled operation");
				break;
			default:
				PyErr_SetString(PyExc_IOError,"Unknown error occured");
				break;
		}
		if(result)
			goto fail;
}
%exception upload_dir_status {
	if(!PyCallable_Check(arg4)) {
			PyErr_SetString(PyExc_TypeError,"Callback must be a callable object");
			goto fail;
	}
	$function
	switch(result) {
			case 0:
				break;
			case -EEXIST:
				PyErr_SetString(PyExc_IOError,"File already exists");
				break;
			case -ENOENT:
				PyErr_SetString(PyExc_IOError,"File or directory does not exist");
				break;
			case -ENOSPC:
				PyErr_SetString(PyExc_IOError,"No space left on device");
				break;
			case IFP_ERR_BAD_FILENAME:
				PyErr_SetString(PyExc_IOError,"Bad filename");
				break;
			case IFP_ERR_USER_CANCEL: /* Should this really be an error?*/
				PyErr_SetString(PyExc_IOError,"User canceled operation");
				break;
			default:
				PyErr_SetString(PyExc_IOError,"Unknown error occured");
				break;
		}
		if(result)
			goto fail;
}

%exception download_dir_status {
	if(!PyCallable_Check(arg4)) {
			PyErr_SetString(PyExc_TypeError,"Callback must be a callable object");
			goto fail;
	}
	$function
	switch(result) {
			case 0:
				break;
			case -EEXIST:
				PyErr_SetString(PyExc_IOError,"File already exists");
				break;
			case -ENOENT:
				PyErr_SetString(PyExc_IOError,"File or directory does not exist");
				break;
			case -ENOSPC:
				PyErr_SetString(PyExc_IOError,"No space left on device");
				break;
			case IFP_ERR_BAD_FILENAME:
				PyErr_SetString(PyExc_IOError,"Bad filename");
				break;
			case IFP_ERR_USER_CANCEL: /* Should this really be an error?*/
				PyErr_SetString(PyExc_IOError,"User canceled operation");
				break;
			default:
				PyErr_SetString(PyExc_IOError,"Unknown error occured");
				break;
		}
		if(result)
			goto fail;
}
/* functions that have a named exception handler above */
/* these are wrappers around the low level functions that make them easier to call from python */
%newobject get_ifp_model;
%newobject get_ifp_device_info;
char *get_ifp_model(struct ifp_device *ifp_dev);
char *get_ifp_device_info(struct ifp_device *ifp_dev);
usb_dev_handle *ifp_find_device(void);
int ifp_finalize(struct ifp_device *ifp_dev);

int upload_file_status(struct ifp_device *ifp_dev,const char *localfile, const char *remotefile,PyObject *py_callback);
int download_file_status(struct ifp_device *ifp_dev,const char *remotefile, const char *localfile,PyObject *py_callback);
int upload_dir_status(struct ifp_device *ifp_dev,const char *localdir, const char *remotedir,PyObject *py_callback);
int download_dir_status(struct ifp_device *ifp_dev,const char *remotedir, const char *localdir,PyObject *py_callback);
/* integer returning functions exception handler.
   An exception occurred if the function returns a negitive number */
%exception {
	if(arg1 == NULL) {
		PyErr_SetString(PyExc_ValueError,"iFP device handle is NULL in function '$name'!");
		goto fail;
	}
	$function
	if(result < 0) {
		char error_message[2048];
		snprintf(error_message,2048,"Error reading information from device in function '$name'. (code %d)",result);
		PyErr_SetString(PyExc_IOError,error_message);
		goto fail;
	}
}

/* device status functions */
int ifp_battery(struct ifp_device *ifp_dev);
int ifp_capacity(struct ifp_device *ifp_dev);
int ifp_freespace(struct ifp_device *ifp_dev);



/* file size functions */
int ifp_read_size(struct ifp_device *ifp_dev);
/* functions that return 0 on success and something else on failure */
%exception {
	/*fprintf(stderr,"Entering function '$name'\n");*/
	if(arg1 == NULL) {
		PyErr_SetString(PyExc_ValueError,"iFP device handle is NULL in function '$name'!");
		goto fail;
	}
	$function
	if(result != 0) {
		char error_message[2048];
		snprintf(error_message,2048,"Error reading information from device in function '$name'. (code %d)",result);
		PyErr_SetString(PyExc_IOError,error_message);
		goto fail;
	}
	/*fprintf(stderr,"Exiting function '$name'\n");*/
}

int ifp_release_device(usb_dev_handle *);

/* format the iFP device
	DOESN'T EXACTLY WORK!
	The device formats, the software returns a USB error and the device hangs. I had to remove the battery, but it seemed to recover fine.
	I'm still missing the VOICE and RECORD directories. Is that a bad thing?
*/
int ifp_format(struct ifp_device *ifp_dev);


/* file naming functions */
int ifp_rename(struct ifp_device *ifp_dev,const char *old_path,const char *new_path);
int ifp_delete(struct ifp_device *ifp_dev,const char *f);
int ifp_mkdir(struct ifp_device *ifp_dev,const char *f);
int ifp_rmdir(struct ifp_device *ifp_dev,const char *f);

/* file reading functions */
int ifp_read_open(struct ifp_device *ifp_dev,const char *f);
int ifp_read_close(struct ifp_device *ifp_dev);
int ifp_read_seek(struct ifp_device *ifp_dev,int abs_position);

%cstring_output_maxsize(void *buf,int bytes);
int ifp_read_data(struct ifp_device *ifp_dev,void *buf, int bytes);

int ifp_read_eof(struct ifp_device *ifp_dev);


/* file writing functions */
int ifp_write_open(struct ifp_device *ifp_dev,const char *f,int fsize);
int ifp_write_close(struct ifp_device *ifp_dev);
%apply (char *STRING, int LENGTH) { (char *b, int bytes) }
int ifp_write_data(struct ifp_device *ifp_dev,char *b,int bytes);

int ifp_delete_dir_recursive(struct ifp_device *ifp_dev,const char *f);

%include "cstring.i"
%cstring_bounded_output(char *callsign_in, IFP_TUNER_LABEL+1);
%apply int *OUTPUT {int *freq};
int get_station_info(struct ifp_device *ifp_dev,int station_num, char *callsign_in, int *freq);
%clear int *OUTPUT;
int set_station_info(struct ifp_device *ifp_dev,int station_num, const char *callsign, int freq);


/* functions that return status in the integer result
   Clear out the exception handler 
*/
%exception;
/* file stat functions */
int ifp_is_file(struct ifp_device *ifp_dev,const char *f);
int ifp_is_dir(struct ifp_device *ifp_dev,const char *f);
int ifp_exists(struct ifp_device *ifp_dev,const char *f);

const char * ifp_error_message (int n);

